"""
Defines the module's version, read from the version.txt file.
"""

import pathlib

with open(
    pathlib.Path(__file__).resolve().parent / "version.txt", encoding="utf-8"
) as f:
    __version__ = f.read().strip()



"""
Defines the (de-)serialization for special built-in types.
"""

from collections.abc import Hashable, Iterable, Mapping
from functools import singledispatch
from numbers import Complex
from types import SimpleNamespace

import numpy as np

from ._base_classes import Deserializable
from ._save_load import from_hdf5, to_hdf5, to_hdf5_singledispatch
from ._subscribe import TYPE_TAG_KEY, subscribe_hdf5

__all__ = []


class _SpecialTypeTags(SimpleNamespace):
    """
    Defines the type tags for special types.
    """

    LIST = "builtins.list"
    TUPLE = "builtins.tuple"
    DICT = "builtins.dict"
    NUMBER = "builtins.number"
    STR = "builtins.str"
    BYTES = "builtins.bytes"
    NONE = "builtins.none"
    NUMPY_ARRAY = "numpy.ndarray"
    SYMPY = "sympy.object"  # defined in _sympy_load.py and _sympy_save.py


@subscribe_hdf5(_SpecialTypeTags.DICT)
class _DictDeserializer(Deserializable):
    """Helper class to de-serialize dict objects."""

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        try:
            items = from_hdf5(hdf5_handle["items"])
            return {_ensure_hashable(k): v for k, v in items}
        # Handle legacy dicts with only string keys:
        except KeyError:
            res = dict()
            value_group = hdf5_handle["value"]
            for key in value_group:
                res[key] = from_hdf5(value_group[key])
            return res


@subscribe_hdf5(_SpecialTypeTags.LIST)
class _ListDeserializer(Deserializable):
    """Helper class to de-serialize list objects."""

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        return _deserialize_iterable(hdf5_handle)


@subscribe_hdf5(_SpecialTypeTags.TUPLE)
class _TupleDeserializer(Deserializable):
    """Helper class to de-serialize tuple objects."""

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        return tuple(_deserialize_iterable(hdf5_handle))


def _deserialize_iterable(hdf5_handle):
    int_keys = [key for key in hdf5_handle if key != TYPE_TAG_KEY]
    return [from_hdf5(hdf5_handle[key]) for key in sorted(int_keys, key=int)]


@subscribe_hdf5(_SpecialTypeTags.NUMBER, extra_tags=(_SpecialTypeTags.BYTES,))
class _ValueDeserializer(Deserializable):
    """Helper class to de-serialize numbers."""

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        return hdf5_handle["value"][()]


@subscribe_hdf5(_SpecialTypeTags.STR)
class _StringDeserializer(Deserializable):
    """Helper class to de-serialize strings."""

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        return hdf5_handle["value"][()].decode("utf-8")


@subscribe_hdf5(_SpecialTypeTags.NUMPY_ARRAY)
class _NumpyArraryDeserializer(Deserializable):
    """Helper class to de-serialize numpy arrays."""

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        if "value" in hdf5_handle:
            return hdf5_handle["value"][()]
        return np.array(_deserialize_iterable(hdf5_handle))


@subscribe_hdf5(_SpecialTypeTags.NONE)
class _NoneDeserializer(Deserializable):
    """Helper class to de-serialize ``None``."""

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        return None


def add_type_tag(tag):
    """
    Decorator which adds the given type tag when creating the HDF5 object.
    """

    def outer(func):
        def inner(obj, hdf5_handle):
            hdf5_handle[TYPE_TAG_KEY] = tag
            func(obj, hdf5_handle)

        return inner

    return outer


@to_hdf5_singledispatch.register(Iterable)
@add_type_tag(_SpecialTypeTags.LIST)
def _(obj, hdf5_handle):
    _serialize_iterable(obj, hdf5_handle)


@to_hdf5_singledispatch.register(tuple)
@add_type_tag(_SpecialTypeTags.TUPLE)
def _(obj, hdf5_handle):
    _serialize_iterable(obj, hdf5_handle)


def _serialize_iterable(obj, hdf5_handle):
    for i, part in enumerate(obj):
        sub_group = hdf5_handle.create_group(str(i))
        to_hdf5(part, sub_group)


@to_hdf5_singledispatch.register(Mapping)
@add_type_tag(_SpecialTypeTags.DICT)
def _(obj, hdf5_handle):
    items_group = hdf5_handle.create_group("items")
    to_hdf5(obj.items(), items_group)


@to_hdf5_singledispatch.register(Complex)
@add_type_tag(_SpecialTypeTags.NUMBER)
def _(obj, hdf5_handle):
    _value_serializer(obj, hdf5_handle)


@to_hdf5_singledispatch.register(str)
@to_hdf5_singledispatch.register(np.str_)
@add_type_tag(_SpecialTypeTags.STR)
def _(obj, hdf5_handle):
    _value_serializer(str(obj), hdf5_handle)


@to_hdf5_singledispatch.register(bytes)
@add_type_tag(_SpecialTypeTags.BYTES)
def _(obj, hdf5_handle):
    _value_serializer(obj, hdf5_handle)


@to_hdf5_singledispatch.register(type(None))
@add_type_tag(_SpecialTypeTags.NONE)
def _(obj, hdf5_handle):
    pass


@to_hdf5_singledispatch.register(np.ndarray)
@add_type_tag(_SpecialTypeTags.NUMPY_ARRAY)
def _(obj, hdf5_handle):  # pylint: disable=missing-docstring
    try:
        _value_serializer(obj, hdf5_handle)
    except TypeError:
        # if the numpy dtype does not have a native HDF5 equivalent,
        # treat it as an iterable instead
        _serialize_iterable(obj, hdf5_handle)


def _value_serializer(obj, hdf5_handle):
    hdf5_handle["value"] = obj


def _ensure_hashable(obj):
    if isinstance(obj, Hashable):
        return obj
    return _to_hashable(obj)


@singledispatch
def _to_hashable(obj):
    raise ValueError(
        f"Cannot convert object '{obj}' of type '{type(obj)}' to a hashable object."
    )


@_to_hashable.register(Iterable)
def _(obj):
    return tuple(_ensure_hashable(val) for val in obj)



"""
Defines free functions to serialize / deserialize bands-inspect objects to HDF5.
"""

from functools import lru_cache, singledispatch

import h5py

from fsc.export import export

from ._subscribe import SERIALIZE_MAPPING, TYPE_TAG_KEY

__all__ = ["save", "load"]


@lru_cache(maxsize=None)
def _get_entrypoint_mapping(group):
    """
    Helper function to get the entry point mapping corresponding to a
    given group name
    """
    import pkg_resources  # pylint: disable=import-outside-toplevel

    return {ep.name: ep for ep in pkg_resources.iter_entry_points(group=group)}


def _try_loading_parts(*, identifier, entry_point_mapping):
    """
    Helper function to load an entrypoint corresponding to the given
    identifier. If an exact match for the identifier is not found, the
    colon-separated "parent identifiers" will be checked.

    For example, for an entry point named 'nobody.expects.the',
    it will check 'nobody.expects.the', 'nobody.expects', 'nobody', in
    that order.

    :param identifier: The identifier for which an entrypoint is to be loaded.
    :type identifier: str

    :param entry_point_mapping: A mapping between entrypoint names and the entrypoints themselves.
    :type entry_point_mapping: dict

    :returns: The name of the entry point that was loaded (if any), or False.
    """
    split_identifier = identifier.split(".")
    for partial_identifier_length in range(len(split_identifier), 0, -1):
        partial_identifier = ".".join(split_identifier[:partial_identifier_length])
        if partial_identifier in entry_point_mapping:
            entry_point_mapping[partial_identifier].load()
            return partial_identifier
    return False


@export
def from_hdf5(hdf5_handle):
    """
    Deserializes the given HDF5 handle into an object.

    :param hdf5_handle: HDF5 location where the serialized object is stored.
    :type hdf5_handle: :py:class:`h5py.File<File>` or :py:class:`h5py.Group<Group>`.
    """
    try:
        type_tag = hdf5_handle[TYPE_TAG_KEY][()].decode("utf-8")
    except KeyError as err:
        raise ValueError(
            f"HDF5 object '{hdf5_handle.name}' cannot be de-serialized: No type information given."
        ) from err
    try:
        obj_class = SERIALIZE_MAPPING[type_tag]
    except KeyError as err:
        partial_tag = _try_loading_parts(
            identifier=type_tag,
            entry_point_mapping=_get_entrypoint_mapping("fsc.hdf5_io.load"),
        )
        if not partial_tag:
            raise KeyError(
                f"Unknown {TYPE_TAG_KEY} '{type_tag}'. The module defining this class has not been imported, and no matching entry point was found."
            ) from err
        try:
            obj_class = SERIALIZE_MAPPING[type_tag]
        except KeyError as err2:
            raise KeyError(
                f"Unknown {TYPE_TAG_KEY} '{type_tag}'. The module defining this class has not been imported, even after loading entry point {partial_tag}."
            ) from err2
    return obj_class.from_hdf5(hdf5_handle)


@export
def to_hdf5(obj, hdf5_handle):
    """
    Serializes a given object to HDF5 format.

    :param obj: Object to serialize.

    :param hdf5_handle: HDF5 location where the serialized object gets stored.
    :type hdf5_handle: :py:class:`h5py.File<File>` or :py:class:`h5py.Group<Group>`.
    """
    if hasattr(obj, "to_hdf5"):
        obj.to_hdf5(hdf5_handle)
    else:
        try:
            to_hdf5_singledispatch(obj, hdf5_handle)
        except SerializerNotFound as exc:
            objtype = type(obj)
            objmodule = objtype.__module__
            if objmodule is None:
                fullname = objtype.__qualname__
            else:
                fullname = objmodule + "." + objtype.__qualname__

            if not _try_loading_parts(
                identifier=fullname,
                entry_point_mapping=_get_entrypoint_mapping("fsc.hdf5_io.save"),
            ):
                raise TypeError(
                    f"Cannot serialize object of type '{fullname}', and no corresponding entry point found."
                ) from exc
            to_hdf5_singledispatch(obj, hdf5_handle)


class SerializerNotFound(TypeError):
    """
    Error to raise when the singledispatch for serializing an object to
    HDF5 format fails to find a matching function.
    """


@export
@singledispatch
def to_hdf5_singledispatch(obj, hdf5_handle):
    """
    Singledispatch function which is called to serialize and object when it does not have a ``to_hdf5`` method.

    :param obj: Object to serialize.

    :param hdf5_handle: HDF5 location where the serialized object gets stored.
    :type hdf5_handle: :py:class:`h5py.File<File>` or :py:class:`h5py.Group<Group>`.
    """
    raise SerializerNotFound(f"Cannot serialize object '{obj}' of type '{type(obj)}'")


@export
def from_hdf5_file(hdf5_file):
    """
    Loads the object from a file in HDF5 format.

    :param hdf5_file: Path of the file.
    :type hdf5_file: str
    """
    with h5py.File(hdf5_file, "r") as f:
        return from_hdf5(f)


load = from_hdf5_file  # pylint: disable=invalid-name
load.__doc__ = """Alias for :func:`from_hdf5_file`."""


@export
def to_hdf5_file(obj, hdf5_file):
    """
    Saves the object to a file, in HDF5 format.

    :param obj: The object to be saved.

    :param hdf5_file: Path of the file.
    :type hdf5_file: str
    """
    with h5py.File(hdf5_file, "w") as f:
        to_hdf5(obj, f)


save = to_hdf5_file  # pylint: disable=invalid-name
save.__doc__ = """Alias for :func:`to_hdf5_file`."""



"""
Module defining the serialization method for sympy objects.
"""

import sympy

from ._special_types import (
    _SpecialTypeTags,
    _value_serializer,
    add_type_tag,
    to_hdf5_singledispatch,
)


@to_hdf5_singledispatch.register(sympy.MatrixBase)
@to_hdf5_singledispatch.register(sympy.Basic)
@add_type_tag(_SpecialTypeTags.SYMPY)
def _(obj, hdf5_handle):
    _value_serializer(sympy.srepr(obj), hdf5_handle)



"""
This module contains functions to save and load objects, using the HDF5 format.
"""

from ._base_classes import *
from ._save_load import *

# Needs to be loaded on import to define the special type serialization.
from ._simple_mapping import *
from ._special_types import *
from ._subscribe import *
from ._version import __version__

# pylint: disable=undefined-variable
__all__ = (
    _save_load.__all__
    + _base_classes.__all__
    + _subscribe.__all__
    + _simple_mapping.__all__
)



"""
Implements a base class for serializing a given list of attributes of an object.
"""

import contextlib

from fsc.export import export

from ._base_classes import HDF5Enabled
from ._save_load import from_hdf5 as _global_from_hdf5
from ._save_load import to_hdf5 as _global_to_hdf5


@export
class SimpleHDF5Mapping(HDF5Enabled):
    """
    Base class for data classes which simply map their member to HDF5 values / groups.

    The child class needs to define a list ``HDF5_ATTRIBUTES`` of attributes which
    should be serialized. The name of the attributes must correspond to the
    name accepted by the constructor.

    For attributes which *can* be serialized but are not required, it can also
    define a list ``HDF5_OPTIONAL``. The same logic as for the ``HDF5_ATTRIBUTES``
    applies, but no error is raised if an attribute does not exist.
    """

    HDF5_ATTRIBUTES = ()
    HDF5_OPTIONAL = ()

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        cls._check_hdf5_attributes_lists()
        kwargs = dict()
        to_deserialize = list(cls.HDF5_ATTRIBUTES) + [
            key for key in cls.HDF5_OPTIONAL if key in hdf5_handle
        ]

        for key in to_deserialize:
            hdf5_obj = hdf5_handle[key]
            try:
                kwargs[key] = hdf5_obj[()]
            except (AttributeError, TypeError):
                kwargs[key] = _global_from_hdf5(hdf5_obj)
        return cls(**kwargs)

    def to_hdf5(self, hdf5_handle):
        self._check_hdf5_attributes_lists()
        to_serialize = [(key, getattr(self, key)) for key in self.HDF5_ATTRIBUTES]
        for key in self.HDF5_OPTIONAL:
            with contextlib.suppress(AttributeError):
                to_serialize.append((key, getattr(self, key)))

        for key, value in to_serialize:
            try:
                hdf5_handle[key] = value
            except TypeError:
                _global_to_hdf5(value, hdf5_handle.create_group(key))

    @classmethod
    def _check_hdf5_attributes_lists(cls):
        """
        Helper method to check that the HDF5_ATTRIBUTES and HDF5_OPTIONAL
        attributes of the class are consistent.
        """
        for key in cls.HDF5_ATTRIBUTES:
            if not isinstance(key, str):
                raise ValueError(
                    f"The element '{key}' in {cls}.HDF5_ATTRIBUTES must be a string."
                )
        for key in cls.HDF5_OPTIONAL:
            if not isinstance(key, str):
                raise ValueError(
                    f"The element '{key}' in {cls}.HDF5_OPTIONAL must be a string."
                )

        overlapping_keys = set(cls.HDF5_ATTRIBUTES) & set(cls.HDF5_OPTIONAL)
        if overlapping_keys:
            raise ValueError(
                "The keys {overlapping_keys} are present in both {cls}.HDF5_ATTRIBUTES and {cls}.HDF5_OPTIONAL".format(
                    overlapping_keys=overlapping_keys, cls=cls
                )
            )



"""
Usage: pip install .[dev]
"""

import sys

from setuptools import setup

PKGNAME = "hdf5_io"
PKGNAME_QUALIFIED = "fsc." + PKGNAME

with open("doc/description.txt") as f:
    DESCRIPTION = f.read().strip()
try:
    with open("doc/README") as f:
        README = f.read()
except OSError:
    README = DESCRIPTION

with open("version.txt") as f:
    VERSION = f.read().strip()

if sys.version_info < (3, 7):
    raise ValueError("only Python 3.7 and higher are supported")

setup(
    name=PKGNAME_QUALIFIED,
    version=VERSION,
    packages=[PKGNAME_QUALIFIED],
    url="http://frescolinogroup.github.io/frescolino/pyhdf5io/"
    + ".".join(VERSION.split(".")[:2]),
    include_package_data=True,
    author="C. Frescolino",
    author_email="frescolino@lists.phys.ethz.ch",
    description=DESCRIPTION,
    install_requires=["numpy>=1.17.5", "decorator", "h5py~=3.0", "fsc.export"],
    python_requires=">=3.7",
    extras_require={
        "dev": [
            "pytest",
            "pytest-cov",
            "pre-commit==2.15.0",
            "pylint==2.11.1",
            "sphinx",
            "sphinx-rtd-theme",
            "ipython>=6.2",
            "matplotlib",
            "sympy",
        ]
    },
    entry_points={
        "fsc.hdf5_io.load": ["sympy.object = fsc.hdf5_io._sympy_load"],
        "fsc.hdf5_io.save": ["sympy = fsc.hdf5_io._sympy_save"],
    },
    long_description=README,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Topic :: Utilities",
    ],
    license="Apache",
)



"""
Module defining the deserialization method for sympy objects.
"""

from ._special_types import Deserializable, _SpecialTypeTags
from ._subscribe import subscribe_hdf5


@subscribe_hdf5(_SpecialTypeTags.SYMPY)
class _SympyDeserializer(Deserializable):
    """Helper class to de-serialize sympy objects."""

    @classmethod
    def from_hdf5(cls, hdf5_handle):
        import sympy  # pylint: disable=import-outside-toplevel

        return sympy.sympify(hdf5_handle["value"][()].decode("utf-8"))
